#include "debug.hpp"
#include "program.hpp"
#include "simulation.h"
#include <GL/glew.h>
#include <GLUT/glut.h>
#include <array>
#include <cmath>
#include <cstdio>
#include <cstdlib>
#include <glm/ext.hpp>
#include <glm/glm.hpp>
#include <sys/stat.h>
#include <vector>

#pragma clang diagnostic ignored "-Wdeprecated-declarations"

bool debugContDist = false;

using namespace std;
using namespace glm;

const float metaballR = 1.5f;
inline float metaballField(float r) {
  if (r > metaballR)
    return 0;
  const float a = r / metaballR;
  return (-4.f / 9.f * powf(a, 6)) + (17.f / 9.f * powf(a, 4)) -
         (22.f / 9.f * powf(a, 2)) + 1;
}

const float normalizationFactor = 748.f / 405.f * M_PI * metaballR;

void checkError() {
  if (GLenum e = glGetError()) {
    fprintf(stderr, "%s\n", gluErrorString(e));
    abort();
  }
}

GLuint bbProg;
GLuint bbVao;

// Here we need to generate n_q textures for different densities of metaballs
// These textures then go on the billboards
// The texture stores attenuation ratio?

#define NQ 64
GLuint bbTexIds[NQ];

// Stores attenuation ratio inside r channel
// Should be highest value at center
void precalculateBillboardTextures() {
  fprintf(stderr, "Calculating billboard textures...\n");
  glGenTextures(NQ, bbTexIds);

  for (int d = 0; d < NQ; d++) {
    float data[32 * 32];
    for (int j = 0; j < 32; j++) {
      for (int i = 0; i < 32; i++) {
        // TODO: properly calculate this instead of whatever this is
        float r = distance(vec2(i, j), vec2(16, 16)) / 16;
        float density = (float)d / NQ;
        data[i + j * 32] =
            1 - (density * metaballField(r) / normalizationFactor);
      }
    }

    mkdir("bbtex", 0777);
    char path[32];
    snprintf(path, 32, "bbtex/%i.tga", d);
    saveGrayscale(data, 32, 32, path);

    glBindTexture(GL_TEXTURE_2D, bbTexIds[d]);
    checkError();

    glTexImage2D(GL_TEXTURE_2D, 0, GL_RED, 32, 32, 0, GL_RED, GL_FLOAT, data);
    glGenerateMipmap(GL_TEXTURE_2D); // required, otherwise texture is blank

    checkError();

    fprintf(stderr, "\r%i out of %i densities calculated%s", d + 1, NQ,
            d == NQ - 1 ? "\n" : "");
  }
}

struct Metaball {
  vec3 pos;
  /** Radius, density */
  float r, d;
  vec4 col;
};

array<Metaball, CLOUD_DIM * CLOUD_DIM * CLOUD_DIM> metaballs;

Clouds cs;

void calculateMetaballs() {
  stepClouds(&cs);
  /* for (int i = 0; i < 256; i++) { */
  /*   float x = ((float)rand()/(float)(RAND_MAX) - 0.5) * 2; */
  /*   float y = ((float)rand()/(float)(RAND_MAX) - 0.5) * 2; */
  /*   float z = ((float)rand()/(float)(RAND_MAX) - 0.5) * 2; */
  /*   float r = (float)rand()/(float)(RAND_MAX) * 1; */
  /*   Metaball m = {{x,y,z}, r}; */
  /*   metaballs.push_back(m); */
  /* } */
  for (int i = 0; i < CLOUD_DIM; i++) {
    for (int j = 0; j < CLOUD_DIM; j++) {
      for (int k = 0; k < CLOUD_DIM; k++) {
        Metaball m = {
            {i / (float)CLOUD_DIM, j / (float)CLOUD_DIM, k / (float)CLOUD_DIM},
            1.f / (float)CLOUD_DIM};
        m.pos = (m.pos * vec3(2)) - vec3(1);
        m.d = cs.q[i][j][k];
        metaballs[i * CLOUD_DIM * CLOUD_DIM + j * CLOUD_DIM + k] = m;
      }
    }
  }
}

vec3 sunPos = {0, 2, 2}, sunDir = {0, -1, -1};
vec3 camPos = {0, 0, -5}, viewPos = {0, 0, 0};
mat4 proj; // projection matrix
mat4 view; // view matrix
float znear = 0.001, zfar = 1000;
float width = 600, height = 400;
float aspect = width / height;

void setProjectionAndViewUniforms(GLuint progId) {
  GLuint projLoc = glGetUniformLocation(progId, "projection");
  glUniformMatrix4fv(projLoc, 1, GL_FALSE, glm::value_ptr(proj));

  GLuint viewLoc = glGetUniformLocation(progId, "view");
  glUniformMatrix4fv(viewLoc, 1, GL_FALSE, glm::value_ptr(view));
}

/** Orientates the transformation matrix to face the camera in the view matrix
 */
mat4 faceView(mat4 m) {
  m[0][0] = view[0][0];
  m[0][1] = view[1][0];
  m[0][2] = view[2][0];
  m[1][0] = view[0][1];
  m[1][1] = view[1][1];
  m[1][2] = view[2][1];
  m[2][0] = view[0][2];
  m[2][1] = view[1][2];
  m[2][2] = view[2][2];
  return m;
}

GLuint attenuationTex;

void shadeClouds() {
  glDisable(GL_DEPTH_TEST);
  // shaderOutput * 0 + buffer * shader alpha
  glBlendFunc(GL_ZERO, GL_SRC_ALPHA);
  glEnable(GL_BLEND);

  // sort by ascending distance from the sun
  sort(metaballs.begin(), metaballs.end(), [](Metaball &a, Metaball &b) {
    return distance(sunPos, a.pos) < distance(sunPos, b.pos);
  });

  glActiveTexture(GL_TEXTURE0);
  glUniform1i(glGetUniformLocation(bbProg, "tex"), 0);

  GLuint modelLoc = glGetUniformLocation(bbProg, "model");

  for (auto &k : metaballs) {
    // place the billboard at the center of k
    mat4 model = scale(translate(mat4(1), k.pos), vec3(k.r) * 2.f);

    // rotate the billboard so that its normal is oriented to the sun
    model = faceView(model);

    glUniformMatrix4fv(modelLoc, 1, GL_FALSE, glm::value_ptr(model));

    // Set the billboard color as RGBA = (1.0, 1.0, 1.0, 1.0).
    vec4 color = {1, 1, 1, 1};
    glUniform4fv(glGetUniformLocation(bbProg, "color"), 1,
                 glm::value_ptr(color));

    // Map the billboard texture with GL_MODULATE.
    // i.e. multiply rather than add
    // but glTexEnv is for the old fixed function pipeline --
    // need to just tell our fragment shader then to modulate
    int dIdx = k.d * NQ;
    glBindTexture(GL_TEXTURE_2D, bbTexIds[dIdx]);
    glUniform1i(glGetUniformLocation(bbProg, "modulate"), 1);

    // Render the billboard.
    glDrawElements(GL_TRIANGLES, 6, GL_UNSIGNED_INT, 0);

    // Read the pixel value corresponding to the center of metaball k.
    // 1. First get position in opengl screen space: from [-1,1]
    // 2. Normalize to [0,1]
    // 3. Multiply by (width * height)
    vec2 screenPos =
        ((vec2(proj * view * model * vec4(0, 0, 0, 1)) + vec2(1)) / vec2(2)) *
        vec2(width, height);
    vec4 pixel;
    // TODO: This is a huge bottleneck
    glReadPixels(screenPos.x, screenPos.y, 1, 1, GL_RGBA, GL_FLOAT,
                 value_ptr(pixel));

    // Multiply the pixel value by the sunlight color.
    vec4 sunColor = {1, 1, 0.9, 1};
    pixel *= sunColor;

    // Store the color into an array C[k] as the color of the billboard.
    k.col = pixel;
  }

  saveFBO();
  checkError();
}

void renderObject() {}

void renderClouds() {
  // Sort metaballs in descending order from the viewpoint
  sort(metaballs.begin(), metaballs.end(), [](Metaball &a, Metaball &b) {
    return distance(camPos, a.pos) > distance(camPos, b.pos);
  });

  glUniform1i(glGetUniformLocation(bbProg, "debugContDist"), debugContDist);

  glDisable(GL_DEPTH_TEST);
  glEnable(GL_BLEND);
  // shaderOutput * 1 + buffer * shader alpha
  glBlendFunc(GL_ONE, GL_SRC_ALPHA);

  glActiveTexture(GL_TEXTURE0);
  glUniform1i(glGetUniformLocation(bbProg, "tex"), 0);

  for (int i = 0; i < metaballs.size(); i++) {
    Metaball k = metaballs[i];

    GLuint modelLoc = glGetUniformLocation(bbProg, "model");

    // Place the billboard at the center of the corresponding metaball n.
    mat4 model = scale(translate(mat4(1), k.pos), vec3(k.r) * 2.f);
    // Rotate the billboard so that its normal is oriented to the viewpoint.
    model = faceView(model);

    glUniformMatrix4fv(modelLoc, 1, GL_FALSE, glm::value_ptr(model));

    // Set the billboard color as C[n].
    k.col.w = 1;
    glUniform4fv(glGetUniformLocation(bbProg, "color"), 1,
                 glm::value_ptr(k.col));

    // Map the billboard texture.
    int dIdx = k.d * NQ;
    glBindTexture(GL_TEXTURE_2D, bbTexIds[dIdx]);

    // Don't modulate it -- blend it
    glUniform1i(glGetUniformLocation(bbProg, "modulate"), 0);

    if (debugContDist) {
      glUniform1f(glGetUniformLocation(bbProg, "density"), k.d);
      glDisable(GL_BLEND);
      model = scale(model, vec3(0.02));
      glUniformMatrix4fv(modelLoc, 1, GL_FALSE, glm::value_ptr(model));
    }

    // Render the billboard with the blending function.
    glDrawElements(GL_TRIANGLES, 6, GL_UNSIGNED_INT, 0);
  }
}

bool needsReshading = true;
void display() {
  if (needsReshading) {
    // TODO: find a way to make sure there's no clipping
    view = glm::lookAt(sunPos + sunDir * vec3(20), sunPos, {0, 1, 0});
    proj = glm::ortho(1.f * aspect, -1.f * aspect, -1.f, 1.f, znear, zfar);
    setProjectionAndViewUniforms(bbProg);

    glClearColor(1, 1, 1, 1);
    glClear(GL_COLOR_BUFFER_BIT);
    shadeClouds();
    needsReshading = false;
  }

  view = glm::lookAt(camPos, viewPos, {0, 1, 0});
  proj = glm::perspective(45.f, aspect, znear, zfar);
  setProjectionAndViewUniforms(bbProg);

  glClearColor(0.83, 1, 1, 1); // background color
  glClear(GL_COLOR_BUFFER_BIT);
  renderObject(); // render things that aren't clouds
  renderClouds();

  glutSwapBuffers();
}

bool needsRedisplay = false;
void timer(int _) {
  /* calculateMetaballs(); */
  if (needsRedisplay) {
    glutPostRedisplay();
    needsRedisplay = false;
  }
  glutTimerFunc(16, timer, 0);
}

void keyboard(unsigned char key, int x, int y) {
  if (key == ' ') {
    calculateMetaballs();
    needsRedisplay = true;
    needsReshading = true;
  }
  if (key == 'd') {
    debugContDist = !debugContDist;
    needsRedisplay = true;
  }
}

int prevMouseX, prevMouseY;
bool firstMouse = true;
void motion(int x, int y) {
  if (firstMouse) {
    prevMouseX = x;
    prevMouseY = y;
    firstMouse = false;
  }
  float dx = x - prevMouseX, dy = y - prevMouseY;
  prevMouseX = x;
  prevMouseY = y;
  const vec3 origin(0, 18, 0);
  const float sensitivity = 0.003f;
  auto camMat = translate(mat4(1), origin + camPos);
  auto rotation = rotate(rotate(mat4(1), -dx * sensitivity, {0, 1, 0}),
                         -dy * sensitivity, {1, 0, 0});
  auto rotAroundOrig = camMat * rotation * translate(mat4(1), origin - camPos);
  camPos = rotAroundOrig * glm::vec4(camPos, 0);
  needsRedisplay = true;
}

void passiveMotion(int x, int y) {
  prevMouseX = x;
  prevMouseY = y;
}

int main(int argc, char **argv) {
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGB |
                      GLUT_3_2_CORE_PROFILE);
  glutInitWindowSize(width, height);
  glutCreateWindow("Clouds");
  glutDisplayFunc(display);

  glewInit();

  Program prog("billboardvert.glsl", "billboardfrag.glsl");

  bbProg = prog.progId;
  glUseProgram(bbProg);

  glGenVertexArrays(1, &bbVao);
  glBindVertexArray(bbVao);
  GLuint vbos[2];
  glGenBuffers(2, vbos);

  vector<vec3> poss = {{-1, -1, 0}, {-1, 1, 0}, {1, 1, 0}, {1, -1, 0}};
  vector<GLuint> indices = {2, 1, 0, 3, 2, 0};

  GLuint posLoc = glGetAttribLocation(bbProg, "vPosition");
  glBindBuffer(GL_ARRAY_BUFFER, vbos[0]);
  glBufferData(GL_ARRAY_BUFFER, poss.size() * sizeof(glm::vec3), &poss[0],
               GL_STATIC_DRAW);
  glEnableVertexAttribArray(posLoc);
  glVertexAttribPointer(posLoc, 3, GL_FLOAT, GL_FALSE, 0, 0);

  glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, vbos[1]);
  glBufferData(GL_ELEMENT_ARRAY_BUFFER, indices.size() * sizeof(GLuint),
               &indices[0], GL_STATIC_DRAW);

  prog.validate();

  precalculateBillboardTextures();

  initClouds(&cs);
  calculateMetaballs();

  glGenTextures(1, &attenuationTex);

  glutKeyboardFunc(keyboard);
  glutMotionFunc(motion);
  glutPassiveMotionFunc(passiveMotion);
  glutTimerFunc(16, timer, 0);

  // set up billboard prog

  glutMainLoop();

  return 0;
}
